indexing
	description: "Objects that map infix and prefix feature names to identifiers"
	author: ""
	date: "$Date$"
	revision: "$Revision$"

class
	CI_OPERATOR_IDENTIFIERS

inherit
	CI_CODE_GENERATION
		export {NONE}
			all
		end

	ET_SHARED_FEATURE_NAME_TESTER

	KL_SHARED_STRING_EQUALITY_TESTER

feature -- Access
	operator_identifier (an_operator: ET_FEATURE_NAME): ET_IDENTIFIER is
			-- Identifier that corresponds to `an_operator'
		require
			is_operator: an_operator /= Void and then (an_operator.is_infix or an_operator.is_prefix)
		local
			str: STRING
			infix_free: ET_INFIX_FREE_NAME
			prefix_free: ET_PREFIX_FREE_NAME
		do
			if an_operator.is_infix_freeop then
				infix_free ?= an_operator
				str := free_infix_code (infix_free)
			elseif an_operator.is_prefix_freeop then
				prefix_free ?= an_operator
				str := free_prefix_code (prefix_free)
			else
				operator_codes.start
				operator_codes.search_forth (an_operator)
				check not operator_codes.after end
				str := operator_codes.key_for_iteration
			end
			create {ET_IDENTIFIER} Result.make (str)
		ensure
			result_exists: Result /= Void
		end

	operator_from_identifier (an_identifier: ET_IDENTIFIER): ET_FEATURE_NAME is
			-- Operator that corresponds to `an_identifier'
			-- Void if `an_identifier' is not valid operator representation
		require
			an_identifier_exists: an_identifier /= Void
		local
			str: STRING
		do
			str := an_identifier.name
			operator_codes.search (str)
			if operator_codes.found then
				Result := operator_codes.found_item
			else
				Result := free_infix_from_code (str)
				if Result = Void then
					Result := free_prefix_from_code (str)
				end
			end
		end

	identifier_name (a_name: ET_FEATURE_NAME): ET_IDENTIFIER is
			-- Identifier that corresponds to `a_name' (if it is already an identifier - `a_name' itself)
		require
			a_name_exists: a_name /= Void
		do
			if a_name.is_identifier then
				Result ?= a_name
			else
				Result := operator_identifier (a_name)
			end
		ensure
			result_exists: Result /= Void
			same_if_identifier: a_name.is_identifier implies Result = a_name
			op_id_if_operator: a_name.is_infix or a_name.is_prefix implies Result.is_equal (operator_identifier (a_name))
		end

feature {NONE} -- Implementation
	infix_operator_prefix: STRING is "infix"
			-- Prefix added to infix operator code

	prefix_operator_prefix: STRING is "prefix"
			-- Prefix added to prefix operator code

	operator_codes: DS_HASH_TABLE [ET_FEATURE_NAME, STRING] is
			-- Identifier-valid code names for standard infix and prefix operators
		once
			create Result.make_with_equality_testers (21, feature_name_tester, string_equality_tester)
			Result.put_new (tokens.infix_and_feature_name, "infix_and")
			Result.put_new (tokens.infix_and_then_feature_name, "infix_and_then")
			Result.put_new (tokens.infix_div_feature_name, "infix_div")
			Result.put_new (tokens.infix_divide_feature_name, "infix_divide")
--			Result.put_new (tokens., "infix_dotdot") -- What about dotdot?
			Result.put_new (tokens.infix_ge_feature_name, "infix_ge")
			Result.put_new (tokens.infix_gt_feature_name, "infix_gt")
			Result.put_new (tokens.infix_implies_feature_name, "infix_implies")
			Result.put_new (tokens.infix_le_feature_name, "infix_le")
			Result.put_new (tokens.infix_lt_feature_name, "infix_lt")
			Result.put_new (tokens.infix_minus_feature_name, "infix_minus")
			Result.put_new (tokens.infix_mod_feature_name, "infix_mod")
			Result.put_new (tokens.infix_or_feature_name, "infix_or")
			Result.put_new (tokens.infix_or_else_feature_name, "infix_or_else")
			Result.put_new (tokens.infix_plus_feature_name, "infix_plus")
			Result.put_new (tokens.infix_power_feature_name, "infix_power")
			Result.put_new (tokens.infix_times_feature_name, "infix_times")
			Result.put_new (tokens.infix_xor_feature_name, "infix_xor")
			Result.put_new (tokens.prefix_minus_feature_name, "prefix_minus")
			Result.put_new (tokens.prefix_not_feature_name, "prefix_not")
			Result.put_new (tokens.prefix_plus_feature_name, "prefix_plus")
		end

	free_infix_code (an_operator: ET_INFIX_FREE_NAME): STRING is
			-- Identifier-valid code name for free infix operator `an_operator'
		require
			an_operator_exists: an_operator /= Void
		do
			Result := free_operator_code (an_operator.operator_name.value, infix_operator_prefix)
		ensure
			result_not_empty: not Result.is_empty
		end

	free_infix_from_code (a_code: STRING): ET_INFIX_FREE_NAME is
			-- Infix operator, who's code name is `a_code'
		require
			a_code_exists: a_code /= Void
		local
			str: STRING
		do
			str := free_operator_from_code (a_code, infix_operator_prefix)
			if str /= Void then
				create Result.make (manifest_string (str))
			end
		end

	free_prefix_code (an_operator: ET_PREFIX_FREE_NAME): STRING is
			-- Identifier-valid code name for free prefix operator `an_operator'
		require
			an_operator_exists: an_operator /= Void
		do
			Result := free_operator_code (an_operator.operator_name.value, prefix_operator_prefix)
		ensure
			result_not_empty: not Result.is_empty
		end

	free_prefix_from_code (a_code: STRING): ET_PREFIX_FREE_NAME is
			-- Prefix operator, who's code name is `a_code'
		require
			a_code_exists: a_code /= Void
		local
			str: STRING
		do
			str := free_operator_from_code (a_code, prefix_operator_prefix)
			if str /= Void then
				create Result.make (manifest_string (str))
			end
		end

	free_operator_code (an_operator, a_kind: STRING): STRING
			-- Code name for operator called `an_operator' of kind `a_kind'
		require
			an_operator_not_empty: an_operator /= Void and then not an_operator.is_empty
			a_kind_is_prefix_or_infix: a_kind /= Void and then
				(a_kind.is_equal (infix_operator_prefix) or a_kind.is_equal (prefix_operator_prefix))
		local
			i: INTEGER
		do
			Result := a_kind.string
			from
				i := 1
			until
				i > an_operator.count
			loop
				Result.append_character ('_')
				Result.append_integer (an_operator.item_code (i))
				i := i + 1
			end
		ensure
			result_not_empty: not Result.is_empty
			inverted_free_operator_from_code: free_operator_from_code (Result, a_kind).is_equal (an_operator)
		end

	free_operator_from_code (a_code, a_kind: STRING): STRING is
			-- Operator name of kind `a_kind' who's code name is `a_code'
			-- Void if `a_code' is not valid code
		require
			a_code_exists: a_code /= Void
			a_kind_is_prefix_or_infix: a_kind /= Void and then
				(a_kind.is_equal (infix_operator_prefix) or a_kind.is_equal (prefix_operator_prefix))
		local
			character_codes: LIST [STRING]
		do
			if a_code.count > a_kind.count + 1 and then a_code.starts_with (a_kind) then
				create Result.make_empty
				character_codes := a_code.substring (a_kind.count + 2, a_code.count).split ('_')
				from
					character_codes.start
				until
					character_codes.after
				loop
					if character_codes.item.is_integer then
						Result.append_character (character_codes.item.to_integer.to_character_8)
					else
						Result := Void
						character_codes.finish
					end
					character_codes.forth
				end
			end
		ensure
			inverted_free_operator_code: Result /= Void implies free_operator_code (Result, a_kind).is_equal (a_code)
		end

invariant
	prefix_operator_prefix_not_empty: not prefix_operator_prefix.is_empty
	infix_operator_prefix_not_empty: not infix_operator_prefix.is_empty
	operator_codes_not_empty: not operator_codes.is_empty
end
