indexing
	description: "Inferred assertions"
	author: ""
	date: "$Date$"
	revision: "$Revision$"

class
	CI_INFERRED_ASSERTION

inherit
	ET_TAGGED_ASSERTION
		rename
			make as make_with_tag
		export {NONE}
			make_with_tag,
			set_expression
		undefine
			tokens
		redefine
			is_equal
		end

	CI_CODE_GENERATION
		export {NONE}
			all
		redefine
			is_equal
		end

create
	make

feature -- Initialization
	make (a_text: STRING) is
			-- Create assertion with text `a_text'
		require
			a_text_exists: a_text /= Void
			a_text_not_empty: not a_text.is_empty
		do
			text := a_text
			create {ET_IDENTIFIER_COLON} tag.make (tag_id, tokens.colon_symbol)
			create text_id.make (text)
			expression := text_id
		ensure
			text_set: text = a_text
		end

feature -- Access
	preconditions: ET_PRECONDITIONS
			-- Conditions under which this assertion can be evaluated
			-- Void, if no preconditions (or they are not known yet)

feature -- Status report
	is_comment: BOOLEAN is
			-- Is this assertion an Eiffel comment?
		do
			Result := text.starts_with (tokens.comment_symbol_name)
		end

	is_current_exists: BOOLEAN is
			-- Is this assertion "Current /= Void"?
		do
			Result := text.has_substring ("Current /= Void")
		end

	is_current_not_changed: BOOLEAN is
			-- Is this assertion "Current = old Current"?
			-- Bug, this assertion shouldn't appear, because Current is decaled with "isParam",
			-- but it appears as conditional assertion. It is printed "Current = Current", because
			-- "old"s are removed for isParam variables
		do
			Result := text.has_substring ("Current = Current")
		end

	is_constant_address: BOOLEAN is
			-- Does this assertion state that address of some variable is constant?
		do
			Result := text.has_substring ("has only one value")
		end

	has_variable (a_variable_name: STRING): BOOLEAN is
			-- Does assertion contain a varibale called `a_variable_name'?
		require
			a_variable_name_not_empty: a_variable_name /= Void and then not a_variable_name.is_empty
		local
			i: INTEGER
		do
			from
				i := text.substring_index (a_variable_name, 1)
			until
				i = 0 or Result
			loop
				Result := is_variable_occurence (i, a_variable_name.count)
				i := text.substring_index (a_variable_name, i + 1)
			end
		end

feature -- Element change
	add_preconditions (a_preconditions: ET_ASSERTIONS) is
			-- Add `a_preconditions' to front of `preconditions'
		require
			a_preconditions_exists: a_preconditions /= Void
		local
			i: INTEGER
		do
			if not a_preconditions.is_empty then
				if preconditions = Void then
					create preconditions.make_with_capacity (a_preconditions.count)
				else
					preconditions.resize (preconditions.count + a_preconditions.count)
				end
				from
					i := a_preconditions.count
				until
					i < 1
				loop
					preconditions.put_first (a_preconditions.item (i))
					i := i - 1
				end
				create {ET_INFIX_EXPRESSION} expression.make (joined (preconditions, tokens.and_then_operator),
					create {ET_KEYWORD_OPERATOR}.make_implies, text_id)
			end
		end

	replace_string_values is
			-- Replace all occurences of a string value variable names with string variable names
		local
			i: INTEGER
			suffix: STRING
		do
			suffix := {CI_STRING_VALUE}.suffix
			from
				i := 1
			until
				i = 0
			loop
				i := text.substring_index (suffix, i)
				if (i > 0) and then is_variable_suffix (i, suffix.count) then
					text.remove_substring (i, i + suffix.count - 1)
				end
			end
		end

feature -- Comparison
	is_equal (other: CI_INFERRED_ASSERTION): BOOLEAN is
			-- Is `other' equal to current?
		do
			Result := text.is_equal (other.text)
		end

feature {CI_INFERRED_ASSERTION} -- Implementation
	text: STRING
			-- Text of assertion

feature {NONE} -- Implementation
	tag_id: ET_IDENTIFIER is
			-- Assertion tag
		once
			create Result.make ("inferred")
		end

	text_id: ET_IDENTIFIER
			-- Identifier with name `text'
			-- Though text isn't a valid identifier, this is the simplest way to print it

	is_variable_occurence (i, nb: INTEGER): BOOLEAN is
			-- Is substring of `text' beginning at `i' with length `nb' an occurence of variable name?
		require
			nb_positive: nb > 0
			i_in_bounds: (i >= 1) and (i <= text.count - nb + 1)
		local
			c1, c2: CHARACTER
		do
			if i > 1 then
				c1 := text.item (i - 1)
			end
			if i + nb - 1 < text.count then
				c2 := text.item (i + nb)
			end
			Result := not tokens.is_identifier_symbol (c1) and not (c1 = tokens.dot_symbol_character) and
				not tokens.is_identifier_symbol (c2) and not (c2 = tokens.dot_symbol_character)
		end

	is_variable_suffix (i, nb: INTEGER): BOOLEAN is
			-- Is substring of `text' beginning at `i' with length `nb' a proper suffix of variable name?
		require
			nb_positive: nb > 0
			i_in_bounds: (i >= 1) and (i <= text.count - nb + 1)
		local
			c1, c2: CHARACTER
		do
			if i > 1 then
				c1 := text.item (i - 1)
			end
			if i + nb - 1 < text.count then
				c2 := text.item (i + nb)
			end
			Result := tokens.is_identifier_symbol (c1) and not tokens.is_identifier_symbol (c2)
		end

invariant
	text_exists: text /= Void
	text_not_empty: not text.is_empty
	tag_id_exists: tag_id /= Void
	text_id_exists: text_id /= Void
end
